from typing import Optional, Dict

from graphql import GraphQLArgument, GraphQLNonNull
from sqlalchemy.ext.declarative import DeclarativeMeta

from .types import get_graphql_type_from_column


def make_pk_args(model: DeclarativeMeta) -> Optional[Dict[str, GraphQLArgument]]:
    primary_key = model.__table__.primary_key  # type: ignore

    if not primary_key:
        return None

    args = {}
    for column in primary_key.columns:
        graphql_type = get_graphql_type_from_column(column)
        args[column.name] = GraphQLArgument(GraphQLNonNull(graphql_type))

    return args



from sqlalchemy import Column, Integer
from graphql import GraphQLInt

from graphql_sqlalchemy.types import get_graphql_type_from_column


def describe_get_graphql_type_from_column():
    def handles_int():
        column = Column("name", Integer)

        assert get_graphql_type_from_column(column) == GraphQLInt



from sqlalchemy import Integer, Float, Boolean
from graphql import GraphQLString, GraphQLInt, GraphQLFloat, GraphQLBoolean


def get_graphql_type_from_column(column):
    if isinstance(column.type, Integer):
        return GraphQLInt

    if isinstance(column.type, Float):
        return GraphQLFloat

    if isinstance(column.type, Boolean):
        return GraphQLBoolean

    return GraphQLString



from typing import Dict

from sqlalchemy import Table
from sqlalchemy.orm import interfaces
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from graphql import (
    GraphQLObjectType,
    GraphQLField,
    GraphQLSchema,
    GraphQLList,
)

from .resolvers import make_field_resolver, make_resolver, make_pk_resolver
from .types import get_graphql_type_from_column
from .args import make_pk_args

Objects = Dict[str, GraphQLObjectType]


def get_model_field_name(model: DeclarativeMeta) -> str:
    return model.__tablename__  # type: ignore


def get_model_pk_field_name(model: DeclarativeMeta) -> str:
    return f"{model.__tablename__}_by_pk"  # type: ignore


def get_table_object_name(table: Table) -> str:
    return table.name.title().replace("_", "")


def build_table_type(model: DeclarativeMeta, objects: Objects) -> GraphQLObjectType:
    table = model.__table__  # type: ignore

    def get_fields():
        fields = {}

        for column in table.columns:
            graphql_type = get_graphql_type_from_column(column)
            fields[column.name] = GraphQLField(graphql_type, resolve=make_field_resolver(column.name))

        for name, relationship in model.__mapper__.relationships.items():
            object_type = objects[relationship.mapper.entity.__tablename__]
            if relationship.direction in (interfaces.ONETOMANY, interfaces.MANYTOMANY):
                object_type = GraphQLList(object_type)

            fields[name] = fields[name] = GraphQLField(object_type, resolve=make_field_resolver(name),)

        return fields

    return GraphQLObjectType(get_table_object_name(table), get_fields)


def build_schema(base: DeclarativeMeta) -> GraphQLSchema:
    fields = {}
    objects: Dict[str, GraphQLObjectType] = {}

    for model in base.__subclasses__():
        field_name = get_model_field_name(model)
        object_type = build_table_type(model, objects)

        objects[field_name] = object_type
        fields[field_name] = GraphQLField(GraphQLList(object_type), resolve=make_resolver(model))

        pk_field_name = get_model_pk_field_name(model)
        fields[pk_field_name] = GraphQLField(object_type, args=make_pk_args(model), resolve=make_pk_resolver(model))

    return GraphQLSchema(GraphQLObjectType("Query", lambda: fields))



from typing import Any, Callable, Dict

from sqlalchemy.ext.declarative import DeclarativeMeta


def make_field_resolver(field: str) -> Callable:
    def resolver(root: DeclarativeMeta, _info: Any):
        return getattr(root, field)

    return resolver


def make_resolver(model: DeclarativeMeta) -> Callable:
    def resolver(_root: DeclarativeMeta, info: Any):
        session = info.context["session"]
        return session.query(model).all()

    return resolver


def make_pk_resolver(model: DeclarativeMeta) -> Callable:
    def resolver(_root: DeclarativeMeta, info: Any, **kwargs: Dict[str, Any]):
        session = info.context["session"]
        return session.query(model).get(kwargs)

    return resolver
